use strict;
use warnings;
use Test::More;

# generated by Dist::Zilla::Plugin::Test::PodSpelling 2.006009
use Test::Spelling 0.12;
use Pod::Wordlist;


add_stopwords(<DATA>);
all_pod_files_spelling_ok( qw( bin lib  ) );
__DATA__
IxHash
JIT
concat
darwin
dec
decrement
decrementing
incrementing
postdec
predec
postinc
preinc
tombstoned
David
Golden
dagolden
Benct
Philip
Jonsson
bpjonsson
Mario
Roy
marioeroy
lib
Hash
Ordered
Benchmarks
